/*
 *  Based on Fldigi source code version 3.21.82
 *
 *
 * Copyright (C) 2014 John Douyere (VK2ETA)
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */
#include "AndFlmsg_Fldigi_interface.h"
//#include "modem.h"
#include "psk.h"
#include "mt63.h"
#include "thor.h"
#include "mfsk.h"
#include "olivia.h"
#include "rsid.h"
#include <jni.h>

using std::string;

modem *active_modem = NULL;
cRsId *RsidModem = NULL;
char decodedBuffer[5000];
int lastCharPos;
JNIEnv* gEnv = NULL;
jobject gJobject = NULL;
//Separate variable for Tx thread as Env are thread sensitive
JNIEnv* txEnv = NULL;
//For Tx buffer handling
signed char* txDataBuffer = NULL;
int txDataBufferLength = 0;
int txCounter = 0;

double audioBuffer[8000];
int audioBufferIndex;


//***************** FROM C++ to JAVA ******************

//Get the value of Java boolean static variable newAmplReady of the Java Modem class
extern bool getNewAmplReady() {
		jclass myClass = NULL;
		myClass = gEnv->FindClass("com/AndFlmsg/Modem");
		jfieldID myFid = NULL;
		//Find Java field newAmplReady of type boolean
		myFid = gEnv->GetStaticFieldID(myClass,"newAmplReady","Z");
	bool newAmplReady = gEnv->GetStaticBooleanField(myClass, myFid);
	return newAmplReady;
return true;
}

//Update the waterfall array of amplitudes in Java class Modem
extern void updateWaterfallBuffer(double *aFFTAmpl) {
	//Find Java Modem clss
	jclass cls = gEnv->FindClass("com/AndFlmsg/Modem");
	//Find Static Java method updateWaterfall (argument is an array of double and returns Void)
	jmethodID mid = gEnv->GetStaticMethodID(cls, "updateWaterfall", "([D)V");
	//Create a Java array and copy the C++ array into it
	jdoubleArray jaFFTAmpl = gEnv->NewDoubleArray(RSID_FFT_SIZE);
	gEnv->SetDoubleArrayRegion(jaFFTAmpl, 0, RSID_FFT_SIZE, aFFTAmpl);
	//Call the method with the Java array
	gEnv->CallStaticVoidMethod(cls, mid, jaFFTAmpl);
	//Release the intermediate array.
	gEnv->DeleteLocalRef(jaFFTAmpl);
}


//Access to Java config class methods for accessing String preferences
extern string getPreferenceS(string preferenceString, string defaultValue) {
	jobject returnedObject;
	jstring jprefStr;
	jstring jdefVal;

	//Check that we have valid data
	if (gEnv ==NULL) return NULL;
	if (gJobject ==NULL) return NULL;

	//Find the Java class
	jclass cls = gEnv->FindClass("com/AndFlmsg/config");

	//Find the static Java method (see signature)
	jmethodID mid = gEnv->GetStaticMethodID(cls, "getPreferenceS", "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;");

	//Convert strings to jstrings
	jprefStr = gEnv->NewStringUTF(preferenceString.c_str());
	jdefVal = gEnv->NewStringUTF(defaultValue.c_str());
	//Call the method
	returnedObject = gEnv->CallStaticObjectMethod(cls, mid, jprefStr, jdefVal);
	//Convert the returned jstring to C string
	const char* str = gEnv->GetStringUTFChars((jstring) returnedObject, NULL);

	//Release the intermediate variable
	gEnv->DeleteLocalRef(returnedObject);
	gEnv->DeleteLocalRef(jprefStr);
	gEnv->DeleteLocalRef(jdefVal);

	//Return the result...that simple...phew
	return str;
}


//Access to Java config class methods for accessing Boolean preferences
extern bool getPreferenceB(string preferenceString, bool defaultValue) {
	jboolean returnedValue;
	jstring jprefStr;

	//Check that we have valid data
	if (gEnv == NULL) return false;
	if (gJobject == NULL) return false;

	//Find the Java class
	jclass cls = gEnv->FindClass("com/AndFlmsg/config");

	//Find the static Java method (see signature)
	jmethodID mid = gEnv->GetStaticMethodID(cls, "getPreferenceB", "(Ljava/lang/String;Z)Z");

	//Convert strings to jstrings
	jprefStr = gEnv->NewStringUTF(preferenceString.c_str());
	jboolean jDefault = defaultValue;
	//Call the method
	returnedValue = gEnv->CallStaticBooleanMethod(cls, mid, jprefStr, jDefault);

	//Release the intermediate variable
	gEnv->DeleteLocalRef(jprefStr);

	//Return the result
	return returnedValue;
}


//Access to Java config class methods for accessing Int preferences
extern double getPreferenceD(string preferenceString, double defaultValue) {
	jdouble returnedValue;
	jstring jprefStr;

	//Check that we have valid data
	if (gEnv == NULL) return 0.0f;
	if (gJobject == NULL) return 0.0f;

	//Find the Java class
	jclass cls = gEnv->FindClass("com/AndFlmsg/config");

	//Find the static Java method (see signature)
	jmethodID mid = gEnv->GetStaticMethodID(cls, "getPreferenceD", "(Ljava/lang/String;D)D");

	//Convert strings to jstrings
	jprefStr = gEnv->NewStringUTF(preferenceString.c_str());
	jdouble jDefault = defaultValue;
	//Call the method
	returnedValue = gEnv->CallStaticDoubleMethod(cls, mid, jprefStr, jDefault);

	//Release the intermediate variable
	gEnv->DeleteLocalRef(jprefStr);

	//Return the result
	return returnedValue;
}


//Access to Java config class methods for accessing Int preferences
extern int getPreferenceI(string preferenceString, int defaultValue) {
	jint returnedValue;
	jstring jprefStr;

	//Check that we have valid data
	if (gEnv == NULL) return 0;
	if (gJobject == NULL) return 0;

	//Find the Java class
	jclass cls = gEnv->FindClass("com/AndFlmsg/config");

	//Find the static Java method (see signature)
	jmethodID mid = gEnv->GetStaticMethodID(cls, "getPreferenceI", "(Ljava/lang/String;I)I");

	//Convert strings to jstrings
	jprefStr = gEnv->NewStringUTF(preferenceString.c_str());
	jint jDefault = defaultValue;
	//Call the method
	returnedValue = gEnv->CallStaticIntMethod(cls, mid, jprefStr, jDefault);

	//Release the intermediate variable
	gEnv->DeleteLocalRef(jprefStr);

	//Return the result
	return returnedValue;
}

//RX Section of the C++ modems
extern void put_rx_char(int receivedChar) {
	decodedBuffer[lastCharPos++] = receivedChar & 0x7f;
}

//----- TX section of the C++ modems ----

//Get one character from the input buffer
extern int get_tx_char() {
	int returnValue;

	returnValue = GET_TX_CHAR_ETX;
	if (txDataBuffer != NULL) {
		if (txCounter < txDataBufferLength) {
			returnValue = (int) (*txDataBuffer++);
			txCounter++;
		}
	}

	return returnValue;
}

//Calls the Java Method that sends the buffer to the sound device
void txModulate(double *buffer, int len) {
	//Accumulate samples
	for (int i = 0; i < len; i++) {
		audioBuffer[audioBufferIndex + i] = buffer[i];
	}
	audioBufferIndex += len;
	if (audioBufferIndex > 2000) {
		//Find the Java class
		jclass cls = txEnv->FindClass("com/AndFlmsg/Modem");
		//Find the static Java method (see signature)
		jmethodID mid = txEnv->GetStaticMethodID(cls, "txModulate", "([DI)V");
		//Create a Java array and copy the C++ array into it
		//debug could this allocation fail?
//		jdoubleArray jBuffer = txEnv->NewDoubleArray(len);
//		txEnv->SetDoubleArrayRegion(jBuffer, 0, len, buffer);
		jdoubleArray jBuffer = txEnv->NewDoubleArray(audioBufferIndex);
		txEnv->SetDoubleArrayRegion(jBuffer, 0, audioBufferIndex, audioBuffer);
		//Call the method with the Java array
//		txEnv->CallStaticVoidMethod(cls, mid, jBuffer, len);
		txEnv->CallStaticVoidMethod(cls, mid, jBuffer, audioBufferIndex);
		//Release the intermediate array and the other variables
		txEnv->DeleteLocalRef(jBuffer);
		txEnv->DeleteLocalRef(cls);
		audioBufferIndex = 0;
	}
}


extern void change_CModem(int modemCode, double newFrequency) {
	//Delete previously created modem to recover memory used
	if (active_modem) {
		delete active_modem;
		active_modem = NULL;
	}
	//Now re-create with new mode
	if (modemCode >= MODE_PSK_FIRST && modemCode <= MODE_PSK_LAST) {
		active_modem = new psk(modemCode);
	} else if (modemCode >= MODE_THOR_FIRST && modemCode <= MODE_THOR_LAST) {
		active_modem = new thor(modemCode);
	} else if (modemCode >= MODE_MFSK_FIRST && modemCode <= MODE_MFSK_LAST) {
		active_modem = new mfsk(modemCode);
	} else if (modemCode >= MODE_MT63_FIRST && modemCode <= MODE_MT63_LAST) {
		active_modem = new mt63(modemCode);
	} else if (modemCode >= MODE_OLIVIA_FIRST && modemCode <= MODE_OLIVIA_LAST) {
		active_modem = new olivia(modemCode);
	}
	//Do the initializations
	if (active_modem != NULL) {
		//First overall modem initilisation
		active_modem->init();
		//Then init of RX side
		active_modem->rx_init();
		//Then set RX frequency
		active_modem->set_freq(newFrequency);
	}
}

//***************** FROM JAVA to C++  ******************

//Save Tx thread environment
extern "C" JNIEXPORT void
Java_com_AndFlmsg_Modem_saveTxEnv( JNIEnv* env, jobject thishere)
{
	txEnv = env;
}


//Creates the RSID receive Modem
extern "C" JNIEXPORT jstring
Java_com_AndFlmsg_Modem_createRsidModem(JNIEnv* env, jobject thishere) {
	//Save environment if we need to call the preference methods in Java
	gEnv = env;
	gJobject = thishere;

	//Delete previously created RSID modem to recover memory used
	if (RsidModem) {
		delete RsidModem;
		RsidModem = NULL;
	}
	RsidModem = new cRsId();
	if (RsidModem != NULL) {
		return (env)->NewStringUTF("RSID Modem created");
	} else {
		return (env)->NewStringUTF("ERROR: RSID Modem Not created");
	}
}


//Returns an array of modem names (uses the RSID list in rsid_def.cxx)
extern "C" JNIEXPORT jobjectArray
Java_com_AndFlmsg_Modem_getModemCapListString(JNIEnv* env,
		jobject thishere)
{
	char *modeCapListString[MAXMODES];
	jobjectArray returnedArray;
	int i;
	int j = 0;

	returnedArray = (jobjectArray)env->NewObjectArray(MAXMODES,
			env->FindClass("java/lang/String"),
			env->NewStringUTF(""));
	for (i=0; i < RsidModem->rsid_ids_size1; i++) {
		if (RsidModem->rsid_ids_1[i].mode != NUM_MODES) {
			env->SetObjectArrayElement(returnedArray,j++,env->NewStringUTF(RsidModem->rsid_ids_1[i].name));
		}
	}

	//do the same for 2nd list
	for (i=0; i < RsidModem->rsid_ids_size2; i++) {
		if (RsidModem->rsid_ids_2[i].mode != NUM_MODES) {
			env->SetObjectArrayElement(returnedArray,j++,env->NewStringUTF(RsidModem->rsid_ids_2[i].name));
		}
	}
//Android debug  env->SetObjectArrayElement(returnedArray,j++,env->NewStringUTF("Bpsk31"));

	//Finalise the array
	env->SetObjectArrayElement(returnedArray,j,env->NewStringUTF("END"));

	return(returnedArray);

}


//Returns an array of modem codes (uses the RSID list in rsid_def.cxx)
extern "C" JNIEXPORT jintArray
Java_com_AndFlmsg_Modem_getModemCapListInt(JNIEnv* env,
		jobject thishere)
{
	jintArray returnedArray;
	int i;
	int j = 0;

	returnedArray = env->NewIntArray(MAXMODES);
	jint temp[MAXMODES];

	for (i=0; i < RsidModem->rsid_ids_size1; i++) {
		if (RsidModem->rsid_ids_1[i].mode != NUM_MODES) {
			temp[j++] = RsidModem->rsid_ids_1[i].mode;
		}
	}

	//do the same for 2nd list
	for (i=0; i < RsidModem->rsid_ids_size2; i++) {
		if (RsidModem->rsid_ids_2[i].mode != NUM_MODES) {
			temp[j++] = RsidModem->rsid_ids_2[i].mode;
		}
	}

	//Finalise the array with a negtive value
	temp[j] = -1;

	//Copy to Java structure
	env->SetIntArrayRegion(returnedArray, 0, MAXMODES, temp);

	return(returnedArray);

}


//Creates the requested modem in C++
extern "C" JNIEXPORT jstring
Java_com_AndFlmsg_Modem_createCModem( JNIEnv* env,
		jobject thishere, jint modemCode)
{
	//Save environment if we need to call the preference methods in Java
	gEnv = env;
	gJobject = thishere;

	//Delete previously created modem to recover memory used
	if (active_modem) {
		delete active_modem;
		active_modem = NULL;
	}

	//Now re-create with new mode
	if (modemCode >= MODE_PSK_FIRST && modemCode <= MODE_PSK_LAST) {
		active_modem = new psk(modemCode);
	} else if (modemCode >= MODE_THOR_FIRST && modemCode <= MODE_THOR_LAST) {
		active_modem = new thor(modemCode);
	} else if (modemCode >= MODE_MFSK_FIRST && modemCode <= MODE_MFSK_LAST) {
		active_modem = new mfsk(modemCode);
	} else if (modemCode >= MODE_MT63_FIRST && modemCode <= MODE_MT63_LAST) {
		active_modem = new mt63(modemCode);
	} else if (modemCode >= MODE_OLIVIA_FIRST && modemCode <= MODE_OLIVIA_LAST) {
		active_modem = new olivia(modemCode);
	} else {
		return (env)->NewStringUTF("ERROR: Modem NOT created");
	}
	return (env)->NewStringUTF("Modem created");
}


//Initializes the RX section of the requested modem in C++
extern "C" JNIEXPORT jstring
Java_com_AndFlmsg_Modem_initCModem( JNIEnv* env,
		jobject thishere, jdouble frequency)
{
	//Save environment if we need to call the preference methods in Java
	gEnv = env;
	gJobject = thishere;

	if (active_modem != NULL) {
		//First overall modem initilisation
		active_modem->init();
		//Then init of RX side
		active_modem->rx_init();
		//Then set RX frequency
		active_modem->set_freq(frequency);
		lastCharPos = 0;
		return (env)->NewStringUTF("Modem Initialized");
	} else {
		return (env)->NewStringUTF("ERROR: Modem NOT Initialized");
	}
}



//Processes the audio buffer through the current modem in C++
extern "C" JNIEXPORT jstring
Java_com_AndFlmsg_Modem_rxCProcess( JNIEnv* env,
		jobject thishere, jshortArray myjbuffer, jint length)
{
	//Save environment if we need to call the preference methods in Java
	gEnv = env;
	gJobject = thishere;

	//Reset received characters buffer pointer
	lastCharPos = 0;
	//Convert to C++ type
	jshort* shortBuffer = env->GetShortArrayElements(myjbuffer, 0);

	//Process the buffer in C++
	if (active_modem != NULL) {
		active_modem->rx_process(shortBuffer, length);
	} else {
		return (env)->NewStringUTF("ERROR: Modem NOT Initialized");
	}
	//Release the passed parameters
	env->ReleaseShortArrayElements(myjbuffer, shortBuffer, 0);

	//Terminate the buffer
	decodedBuffer[lastCharPos] = 0;

	//Return the decoded data
	return env->NewStringUTF(decodedBuffer);
}


//Updates the modem with the latest GUI squelch level
extern "C" JNIEXPORT jint
Java_com_AndFlmsg_Modem_setSquelchLevel( JNIEnv* env,
		jobject thishere, jdouble squelchLevel)
{
	if ( active_modem != NULL) {
			active_modem->set_squelchLevel(squelchLevel);
		}
	return 0;
}


//Returns the latest signal metric from the current modem in C++
extern "C" JNIEXPORT jdouble
Java_com_AndFlmsg_Modem_getMetric( JNIEnv* env,
		jobject thishere)
{
	double metric = 0.0;
	if ( active_modem != NULL) {
		metric = active_modem->get_metric();
	}
	return metric;
}

//Processes the audio buffer through the RSID modem
extern "C" JNIEXPORT jstring
Java_com_AndFlmsg_Modem_RsidCModemReceive( JNIEnv* env,
		jobject thishere, jfloatArray myfbuffer, jint length, jboolean doSearch)
{

	//Convert to C++ type
	jboolean myjcopy = true;
	jfloat* floatBuffer = env->GetFloatArrayElements(myfbuffer, &myjcopy);
	//Reset received characters buffer pointer
	lastCharPos = 0;
	//Process the buffer in C++
	RsidModem->receive(floatBuffer, length, doSearch);
	//Release the passed parameters
	env->ReleaseFloatArrayElements(myfbuffer, floatBuffer, 0);
	//Terminate the buffer
	decodedBuffer[lastCharPos] = 0;

	return env->NewStringUTF(decodedBuffer);

}

//Returns the current modem frequency (for waterfall tracking after RSID)
extern "C" JNIEXPORT jdouble
Java_com_AndFlmsg_Modem_getCurrentFrequency( JNIEnv* env,
		jobject thishere)
{
	return active_modem->get_freq();
}

//Returns the current mode (for updating the Java side)
extern "C" JNIEXPORT jint
Java_com_AndFlmsg_Modem_getCurrentMode( JNIEnv* env,
		jobject thishere)
{
	return active_modem->get_mode();
}

//Returns the decoded characters resulting from the flushing of the rx pipe on RSID rx of new modem
extern "C" JNIEXPORT jint
Java_com_AndFlmsg_Modem_getFlushedRxCharacters( JNIEnv* env,
		jobject thishere)
{
	return active_modem->get_mode();
}




//Transmit section-------------------


//Processes the data buffer for TX through the current modem in C++
extern "C" JNIEXPORT void
Java_com_AndFlmsg_Modem_txRSID(JNIEnv* env, jobject thishere)
{
	if (RsidModem != NULL && active_modem != NULL) {
		RsidModem->send(true); //Always true as we handle post-rsid decision in higher level
	}
}


//Initializes the TX section of the requested modem in C++
extern "C" JNIEXPORT jstring
Java_com_AndFlmsg_Modem_txInit( JNIEnv* env,
		jobject thishere, jdouble frequency)
{
	//Init static variable
	audioBufferIndex = 0;

	if (active_modem != NULL) {
		//Then set TX frequency
		active_modem->set_freq(frequency);
		//Init of TX side
		active_modem->tx_init();
		return (env)->NewStringUTF("Tx Modem Initialized");
	} else {
		return (env)->NewStringUTF("ERROR: Tx Modem NOT Initialized");
	}
}



//Processes the data buffer for TX through the current modem in C++
extern "C" JNIEXPORT jboolean
Java_com_AndFlmsg_Modem_txCProcess( JNIEnv* env,
		jobject thishere, jbyteArray myjbuffer, jint length)
{
	//Reset the static variables
	txCounter = 0;
	//Convert to C++ type
	txDataBuffer = env->GetByteArrayElements(myjbuffer, 0);
	txDataBufferLength = length;
	//Process the buffer in C++
	if (active_modem != NULL) {
		//As there is no idle time in Flmsg context, keep
		//  processing until we have exhausted the data buffer
		while (active_modem->tx_process() >= 0){
			//Processing is done in the test
		}
	} else {
		return JNI_FALSE;
	}
	//Release the passed parameters
	env->ReleaseByteArrayElements(myjbuffer, txDataBuffer, 0);

	//Return the status
	return JNI_TRUE;
}


