/*
 * Message.java  
 *
 * Copyright (C) 2014 John Douyere (VK2ETA) 
 *
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
 * GNU General Public License for more details.  
 *
 * You should have received a copy of the GNU General Public License  
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 *
 */


package com.AndFlmsg;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import android.content.Intent;
import android.net.Uri;


public class Message {

	public static String formName = new String();
	public static String fullDataString = new String();
	
	private final static String radiogramPrecedence[] ={"ROUTINE", "WELFARE", "PRIORITY", "EMERGENCY", "TEST ROUTINE", "TEST WELFARE", "TEST PRIORITY", "TEST EMERGENCY"};
	private final static String iaruPrecedence[] ={"ROUTINE", "PRIORITY", "EMERGENCY"};

	//Declaration of native classes
	public native static 	void 	saveEnv();
    public native static 	boolean ProcessWrapBuffer(String rawWrapBuffer);
    public native static 	String	getUnwrapText();
    public native static 	String	getUnwrapFilename();
    public native static 	String	geterrtext();
    public native static  	String	customHtmlDisplayFormat(String html, String dataPairs);
    public native static  	String	customHtmlEditingFormat(String html, String dataPairs);
    public native static  	String	customHtmlEditFormat(String html, String dataPairs);
    public native static  	String	namedFile();
    public native static  	String	createCustomBuffer(String myHeader, int reason, String dataBuffer);
    public native static  	String	createHardCodedHeader(String myHeader, int reason, String myFormName);
    public native static	String	compressMaybe(String outDataBuffer);
    public native static  	String	updateHeader(String myHeader, int reason, String headerForm);
    public native static 	String	estimate(String jModeName, String jBufferString);
    public native static 	String	dateTimeStamp();
    public native static 	String	dateStamp();
    public native static 	String	timeStamp();
    public native static 	String[] getArlMsgTextList();
    public native static 	String[] getArlHxTextList();
    public native static 	String	cbrgcheck(String msgFieldValue, boolean stdFormat);
    public native static 	String	gettxtrgmsg();
    public native static 	String	escape(String stringToEscape);
    public native static 	String	unescape(String stringToUnescape);
    public native static 	String	expandarl(String msgToExpand);

       
    static {
    	System.loadLibrary("AndFlmsg_Flmsg_Interface");
	}

	public static void addEntryToLog(String entry)
	{
		String logFileName = Processor.HomePath + Processor.Dirprefix + Processor.DirLogs + 
				Processor.Separator + Processor.messageLogFile;
		File logFile = new File(logFileName);
		if (!logFile.exists())
		{
			try
			{
				logFile.createNewFile();
			} 
			catch (IOException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		try
		{
			//BufferedWriter for performance, true to set append to file flag
			BufferedWriter buf = new BufferedWriter(new FileWriter(logFile, true)); 
			buf.append(entry);
			buf.newLine();
			buf.flush();
			buf.close();
		}
		catch (IOException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	

	public static boolean saveMessageAsFile(String filePath, String fileName, String dataString) {
		FileWriter out = null;
		File fileToSave = new File(filePath + fileName);
		try
		{
			if (fileToSave.exists()) {
				fileToSave.delete();
			}
			out = new FileWriter(fileToSave, true);
			out.write(dataString);
			out.close();
			//Update the serial number used for the file name
			if (config.getPreferenceB("SERNBR_FNAME", true)) {
				int serNbr = config.getPreferenceI("SERNBR", 1);
				config.setPreferenceS("SERNBR", Integer.toString(++serNbr));
			}
			AndFlmsg.topToastText("\nSaved file: " + fileName + "\n");
			addEntryToLog(Message.dateTimeStamp() + ": Saved Message file " + fileName);
		}
		catch (IOException e)
		{
			loggingclass.writelog("Error creating file", e, true);
			return false;
		}
		return true;
	}

	
	//Same as saveMessageAsFile but don't increment the serial number counter
	public static boolean saveDraftAsFile(String filePath, String fileName, String dataString) {
		FileWriter out = null;
		File fileToSave = new File(filePath + fileName);
		try
		{
			if (fileToSave.exists()) {
				fileToSave.delete();
			}
			out = new FileWriter(fileToSave, true);
			out.write(dataString);
			out.close();
			AndFlmsg.topToastText("\nSaved file: " + fileName + "\n");
			addEntryToLog(Message.dateTimeStamp() + ": Saved Draft file " + fileName);
		}
		catch (IOException e)
		{
			loggingclass.writelog("Error creating file", e, true);
			return false;
		}
		return true;
	}
	

	//Delete file 
	public static boolean deleteFile(String mFolder, String fileName, boolean adviseDeletion) {

		String fullFileName = Processor.HomePath + Processor.Dirprefix + mFolder + Processor.Separator + fileName;
		File n = new File(fullFileName);
		if (!n.isFile()) {
			return false;
		} else {
			n.delete();
			if (adviseDeletion) {
				AndFlmsg.middleToastText("Document Deleted...");
				addEntryToLog(Message.dateTimeStamp() + ": Deleted file " + fileName);
			}
			return true;
		}
	}

	

	//Copy binary or text files from one folder to another
	public static boolean copyAnyFile(String originFolder, String fileName, String destinationFolder, boolean adviseCopy) {

		File dir = new File(Processor.HomePath + Processor.Dirprefix + destinationFolder);
		if (dir.exists()) {
			String fullFileName = Processor.HomePath + Processor.Dirprefix + originFolder + Processor.Separator + fileName;
			File mFile = new File(fullFileName);
			if (!mFile.isFile()) {
				return false;
			} else {
				FileOutputStream fileOutputStrm = null;
				FileInputStream fileInputStrm = null;
				try {
					fileInputStrm = new FileInputStream(fullFileName);
					String fullDestinationFileName = Processor.HomePath + Processor.Dirprefix + destinationFolder + Processor.Separator + fileName;
					fileOutputStrm = new FileOutputStream(fullDestinationFileName);
					byte[] mBytebuffer = new byte[256];
					int byteCount = 0;
					while ((byteCount = fileInputStrm.read(mBytebuffer)) != -1) {
						fileOutputStrm.write(mBytebuffer, 0, byteCount);
					}
				}
				catch (FileNotFoundException e) {
					Processor.PostToTerminal("File not found: " + fullFileName + "\n");
				}
				catch (IOException e) {
					Processor.PostToTerminal("Error copying: " + fileName + " " + e + "\n");
				}
				finally {
					try {
						if (fileInputStrm != null) {
							fileInputStrm.close();
						}
						if (fileOutputStrm != null) {
							fileOutputStrm.close();
						}
					}
					catch (IOException e) {
						Processor.PostToTerminal("File close error: " + e + "\n");
					}
				}
				if (adviseCopy) 
					addEntryToLog(Message.dateTimeStamp() + ": Copied file " + fileName + " to " + destinationFolder);
				return true;
			}
		} else {
			Processor.PostToTerminal("Directory not found: " + destinationFolder + "\n");
			return false;
		}
	}


	//Copy binary or text files from one folder to another
	public static boolean copyAnyFile(String originFolder, String fileName, 
			String destinationFolder, String newFileName) {
		boolean result = true;
		
		File dir = new File(Processor.HomePath + Processor.Dirprefix + destinationFolder);
		if (dir.exists()) {
			String fullFileName = Processor.HomePath + Processor.Dirprefix + originFolder + Processor.Separator + fileName;
			File mFile = new File(fullFileName);
			if (!mFile.isFile()) {
				result = false;
			} else {
				FileOutputStream fileOutputStrm = null;
				FileInputStream fileInputStrm = null;
				try {
					fileInputStrm = new FileInputStream(fullFileName);
					String fullDestinationFileName = Processor.HomePath + Processor.Dirprefix + destinationFolder + Processor.Separator + newFileName;
					fileOutputStrm = new FileOutputStream(fullDestinationFileName);
					byte[] mBytebuffer = new byte[256];
					int byteCount = 0;
					while ((byteCount = fileInputStrm.read(mBytebuffer)) != -1) {
						fileOutputStrm.write(mBytebuffer, 0, byteCount);
					}
				}
				catch (FileNotFoundException e) {
					Processor.PostToTerminal("File not found: " + fullFileName + "\n");
					result = false;
				}
				catch (IOException e) {
					Processor.PostToTerminal("Error copying: " + fileName + " " + e + "\n");
					result = false;
				}
				finally {
					try {
						if (fileInputStrm != null) {
							fileInputStrm.close();
						}
						if (fileOutputStrm != null) {
							fileOutputStrm.close();
						}
					}
					catch (IOException e) {
						Processor.PostToTerminal("File close error: " + e + "\n");
						result = false;
					}
				}
			}
		} else {
			Processor.PostToTerminal("Directory not found: " + destinationFolder + "\n");
			result = false;
		}
		if (result) addEntryToLog(Message.dateTimeStamp() + ": Copied file " + newFileName + " to " + destinationFolder);
		return result;
	}



	//Creates a temtorary file in Temp folder and a share intent (ready for a start Activity statement)
	//Returns the Intent. Used for sharing the form over the internet (Cloud, email, instant messaging, etc...)
	public static Intent shareInfoIntent(String mDisplayForm, String mFileName, String extension) {

		FileWriter out = null;
		String tempFileName = new String();
		//Create temporary html file to attach to email
		tempFileName = Processor.HomePath + 
				Processor.Dirprefix + Processor.DirTemp + Processor.Separator + 
				mFileName.substring(0,mFileName.lastIndexOf(".")) + extension;
		//Debug
		//Processor.PostToTerminal("tempFileName <" + tempFileName + ">\n\n");
		File htmlFormatedFile = new File(tempFileName);
		try
		{
			if (htmlFormatedFile.exists()) {
				htmlFormatedFile.delete();
			}
			out = new FileWriter(htmlFormatedFile, true);
			out.write(mDisplayForm);
			out.close();
		}
		catch (IOException e)
		{
			loggingclass.writelog("Error creating temporary file in Temp folder", e, true);
		}

		Intent share = new Intent(Intent.ACTION_SEND);
		share.setType("text/plain");
		share.putExtra(Intent.EXTRA_TEXT, "Form attached\n");
		//share.putExtra(Intent.EXTRA_HTML_TEXT, myDisplayForm);
		//share.putExtra(Intent.EXTRA_TEXT, mDisplayForm);
		share.putExtra(Intent.EXTRA_STREAM, Uri.fromFile(new File(tempFileName)));
		//startActivity(Intent.createChooser(share, "Share Form"));
		//Intent emailIntent = new Intent(Intent.ACTION_SENDTO, 
		//								Uri.fromParts("mailto", mailId, null)); 
		share.putExtra(Intent.EXTRA_SUBJECT, "Message " + mFileName); 
		// you can use simple text like this
		// emailIntent.putExtra(android.content.Intent.EXTRA_TEXT,"Body text here"); 
		// or get fancy with HTML like this
		// share.putExtra(Intent.EXTRA_TEXT, Html.fromHtml(myDisplayForm));
		return share;
	}
	
	
	//Takes a directory for the data file, a data file name, and an extension for the form file name and returns a formatted string
	//Returns an HTML format String. For custom forms, ALL INPUT field elements are marked as read-only
	public static String formatForDisplay(String mDir, String mFileName, String extension) {
		boolean isCustomForm = false;
		String dataString = "";
		String mDisplayForm = "";
		fullDataString = "";
		
		try
		{
			//First determine the file name of the form used for displaying the data
			//For this, read the data file until we find the form information
			File fi = new File(Processor.HomePath + Processor.Dirprefix
					+ mDir + Processor.Separator + mFileName);
			if (!fi.isFile()) {
				AndFlmsg.middleToastText("Data file " + mFileName + " Not Found. Deleted/Moved?\n");
				return "";
			}
			FileInputStream fileISi = new FileInputStream(fi);
			BufferedReader buf0 = new BufferedReader(new InputStreamReader(fileISi));
			String readString0 = new String();
			boolean found_flmsg = false;
			boolean found_form = false;
			isCustomForm = false;
			dataString = "";
			
			//Handles both hard-coded forms and custom forms
			while ((readString0 = buf0.readLine()) != null)
			{
				//While we are reading the file, store its complete content in a string for future usage
				dataString += readString0 + "\n";
				fullDataString += readString0 + "\n";
				if (isCustomForm && readString0.contains("CUSTOM_FORM")) {
					//Overwrite formName with custom form file name
					//For custom form, include the extension as it is variable
					formName = readString0.substring(readString0.indexOf("CUSTOM_FORM") + 12);
					found_form = true;
					//Reset the dataString to start at content of :mg: field (CUSTOM_FORM ... then pairs of name, values)
					dataString = readString0.substring(readString0.indexOf("CUSTOM_FORM")) + "\n";
				}
				if (found_flmsg && readString0.contains(">") && readString0.startsWith("<")) {
					found_form = true;
					formName = readString0.substring(1, readString0.lastIndexOf('>')) + extension;
					//Reset dataString so that we excude the header information
					dataString = "";
				}
				if (found_form && formName.equals("customform.html")) {
					isCustomForm = true;
					found_form = false; //Re-set until we found the custom form name
				}
				if (readString0.startsWith("<flmsg>")) {
					found_flmsg = true;
				} 
			}
			//Include error handling if form not valid or does not include form name
			if (!found_flmsg || !found_form){
				AndFlmsg.middleToastText("No form name found in message file or missing <flmsg> statement at start of file. Is it corrupted?");
				return "";
			}
			//First read the form file into a string variable
			//Debug  Processor.PostToTerminal("Templ. name <" + formName + ">\n\n");
			File f = new File(Processor.HomePath + Processor.Dirprefix
					+ Processor.DirDisplayForms + Processor.Separator + formName);
			FileInputStream fileIS = new FileInputStream(f);
			BufferedReader buf = new BufferedReader(new InputStreamReader(fileIS));
			String readString = new String();
			while ((readString = buf.readLine()) != null)
			{
				mDisplayForm += readString + "\n";
			}
			//Then two processings here: standard hard-coded forms and Custom forms
			if (isCustomForm) {
				//Call Native code to format in HTML with updated VALUE and ReadOnly attributes
				mDisplayForm = Message.customHtmlDisplayFormat(mDisplayForm, dataString); //C++ code
				//Debug
				//Processor.PostToTerminal("\nReturned from C++ <" + mDisplayForm + ">\n\n");
			} else { //Hard coded forms format
				//Perform form specific processing
				if (formName.startsWith("radiogram")) {
					//Build and add the :rx: field from preferences
					String rxField = config.getPreferenceS("CALL", "") + "<br>" +
							config.getPreferenceS("TEL", "") + "<br>" +
							config.getPreferenceS("NAME", "") + "<br>" +
							config.getPreferenceS("ADDRESS", "") + "<br>" +
							config.getPreferenceS("CITYSTATEZIP", "")+ "<br>" +
							config.getPreferenceS("EMAIL", "");
					dataString += ":rx:" + Integer.toString(rxField.length()) + " " + rxField + "\n";
					//If selected, expand the ARL Messages
					if (config.getPreferenceB("SHOWARLDESC", true)) {
						String messageField = getHcfFieldContent(dataString, "msg");
						String expandedStr = expandarl(messageField);
						dataString += ":exp:" + Integer.toString(expandedStr.length()) + " " + expandedStr + "\n";
					} 
					//Reformat the boolean variables
					if (dataString.indexOf(":std:1 T") != -1) dataString = dataString.replace(":std:1 T", ":std:2 ON");
					if (dataString.indexOf(":std:1 F") != -1) dataString = dataString.replace(":std:1 F", ":std:3 OFF");
					if (dataString.indexOf(":svc:1 T") != -1) dataString = dataString.replace(":svc:1 T", ":svc:2 ON");
					if (dataString.indexOf(":svc:1 F") != -1) dataString = dataString.replace(":svc:1 F", ":svc:3 OFF");
					//Reformat the "precedence" field
					int precValue = Integer.parseInt(getHcfFieldContent(dataString, "prec"));
					if (precValue < 0 || precValue >= radiogramPrecedence.length) precValue = 0;
					dataString = setHcfFieldContent(dataString, "prec", radiogramPrecedence[precValue]);
				} else if (formName.startsWith("iaru")) {
					//Reformat the "precedence" field
					int precValue = Integer.parseInt(getHcfFieldContent(dataString, "prec"));
					if (precValue < 0 || precValue >= iaruPrecedence.length) precValue = 0;
					dataString = setHcfFieldContent(dataString, "prec", iaruPrecedence[precValue]);
				} else if (formName.startsWith("ics206")) {
					//Reformat the boolean variables
					for (int i=0; i <5; i++) {
						if (dataString.indexOf(":mpar" + Integer.toString(i) + ": T") != -1) dataString = dataString.replace(":mpar" + Integer.toString(i) + ": T", ":mpar" + Integer.toString(i) + ":3 YES");
						if (dataString.indexOf(":tpara" + Integer.toString(i) + ": T") != -1) dataString = dataString.replace(":tpara" + Integer.toString(i) + ": T", ":tpara" + Integer.toString(i) + ":3 YES");
						if (dataString.indexOf(":apara" + Integer.toString(i) + ": T") != -1) dataString = dataString.replace(":apara" + Integer.toString(i) + ": T", ":apara" + Integer.toString(i) + ":3 YES");
						if (dataString.indexOf(":hheli" + Integer.toString(i) + ": T") != -1) dataString = dataString.replace(":hheli" + Integer.toString(i) + ": T", ":hheli" + Integer.toString(i) + ":3 YES");
						if (dataString.indexOf(":hburn" + Integer.toString(i) + ": T") != -1) dataString = dataString.replace(":hburn" + Integer.toString(i) + ": T", ":hburn" + Integer.toString(i) + ":3 YES");
					}
				}
				//Perform standard processing for Hard Coded forms
				//Send file name as :TITLE: field
				dataString += ":TITLE:" + Integer.toString(mFileName.length()) + " " + mFileName + "\n";
				//Scan the formatted dataString for field and their values 
				//    (The 3 patterns are IN THAT ORDER of MATCH <:field:length value> OR <\n> OR <any Text>)
				Pattern psc = Pattern.compile("^(:.+:)(\\d+) (.*)|(^\\n)|(.+)", Pattern.MULTILINE);
				Matcher msc = psc.matcher(dataString);
				boolean keepLooking = true;
				String element = "";
				String value = "";
				String tempBuffer = "";
				int fieldLength = 0;
				int runningFieldLength = 0;
				boolean newUncompleteField = false;
				//If a match is found, ensure we read the <Length> number of characters before declaring we have a complete field.
				// Replace the fields in the form with the values found as we go.
				for (int start = 0; (keepLooking && start < dataString.length());)
				{
					keepLooking = msc.find(start);
					if (keepLooking) {
						if (msc.group(1) != null) { //We have a ":fieldname:nn fieldvalue" line
							newUncompleteField = true;//Assumed true, reset below if not true
							element = msc.group(1);
							fieldLength = Integer.parseInt(msc.group(2));
							value = msc.group(3);
							if (fieldLength <= value.length()) {
								newUncompleteField = false;
								mDisplayForm = mDisplayForm.replace(element, value);
							} else {
								runningFieldLength = value.length();
								tempBuffer = value;
							}
						} else if (msc.group(4) != null) {//We have just a line feed, nothing else
							if (newUncompleteField) {
								String a4 = msc.group(4);
								runningFieldLength += a4.length();
								tempBuffer += a4;
								if (fieldLength <= runningFieldLength) {
									newUncompleteField = false;
									mDisplayForm = mDisplayForm.replace(element, tempBuffer);
									tempBuffer = "";
								}
							}
						} else if (msc.group(5) != null){ //We have a 2nd, 3rd etc.. line of text
							if (newUncompleteField) {
								value = msc.group(5);
								runningFieldLength += value.length() + 1;
								tempBuffer += "\n" + value;
								if (fieldLength <= runningFieldLength) {
									newUncompleteField = false;
									//tempBuffer += "\"\n";
									mDisplayForm = mDisplayForm.replace(element, tempBuffer);
									tempBuffer = "";
								}
							}
						}
					}
					if (keepLooking) start = msc.end();
				}
				//Then Remove the missing ":xxxx:" field templates that were not transmitted (because they were blank / Un-checked)
				String fieldName = new String();
				psc = Pattern.compile("\\:\\S+\\:");
				msc = psc.matcher(mDisplayForm);
				while (msc.find())
				{
					fieldName = msc.group(0);
					mDisplayForm = mDisplayForm.replace(fieldName, "");
				}
				//Finally, if it is for a text display replace all \n with <br> as we display in a Webview
				if (extension.equals(".txt")) {
					mDisplayForm = mDisplayForm.replaceAll("\n", "<br>");
				}
			}
		}
		catch (FileNotFoundException e)
		{
			AndFlmsg.middleToastText("Form file " + formName + " Not Found. Can't format this message, displaying raw file");
			//Convert the HTML reserved characters "<" and ">" to their entity name
			mDisplayForm = fullDataString.replaceAll("<", "&lt");
			mDisplayForm = mDisplayForm.replaceAll(">", "&gt");
			//Since it is displayed as text, replace all \n with <br> as we display in a Webview
			mDisplayForm = mDisplayForm.replaceAll("\n", "<br>");
			//Then escape the special characters
			mDisplayForm = escape(mDisplayForm);
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		return mDisplayForm;
	}

	
	//Takes a directory for the data file, a data file name, and returns a formatted string
	//Returns an HTML format String. For custom forms, ALL field elements are marked as read-only
	public static String formatForEditing(String mDir, String mFileName) {
		boolean isCustomForm = false;
		String dataString = new String();
		String mDisplayForm = new String();
		
		try
		{
			//First determine the file name of the form used for displaying the data
			//For this read the data file until we find the form information
			File fi = new File(Processor.HomePath + Processor.Dirprefix
					+ mDir + Processor.Separator + mFileName);
			FileInputStream fileISi = new FileInputStream(fi);
			BufferedReader buf0 = new BufferedReader(new InputStreamReader(fileISi));
			String readString0 = new String();
			boolean found_flmsg = false;
			boolean found_form = false;
			isCustomForm = false;
			dataString = "";
			//Handles both hard-coded forms and custom forms
			while ((readString0 = buf0.readLine()) != null)
			{
				//While we are reading the file, store its complete content in a string for future usage
				dataString += readString0 + "\n";
				if (isCustomForm && readString0.contains("CUSTOM_FORM")) {
					//Overwrite formName with custom form file name
					//For custom form, include the extension as it is variable
					formName = readString0.substring(readString0.indexOf("CUSTOM_FORM") + 12);
					found_form = true;
					//Reset the dataString to start at content of :mg: field (CUSTOM_FORM ... then pairs of name, values)
					dataString = readString0.substring(readString0.indexOf("CUSTOM_FORM")) + "\n";
				}
				if (found_flmsg && readString0.contains(">") && readString0.startsWith("<")) {
					found_form = true;
					formName = readString0.substring(1, readString0.lastIndexOf('>')) + ".html";
					//Save header if we save draft message
					AndFlmsg.mHeader = dataString;
					dataString = ""; //Reset dataString
				}
				if (found_form && formName.equals("customform.html")) {
					isCustomForm = true;
					found_form = false; //Re-set until we found the custom form name
				}
				if (readString0.startsWith("<flmsg>")) {
					found_flmsg = true;
				} 
			}
			//Include error handling if form not valid or does not include form name
			if (!found_flmsg || !found_form){
				AndFlmsg.middleToastText("No form name found in message file or missing <flmsg> statement at start of file. Is it corrupted?");
				return "";
			}
			//First read the form file into a string variable
			//Debug  Processor.PostToTerminal("Templ. name <" + formName + ">\n\n");
			File f = new File(Processor.HomePath + Processor.Dirprefix
					+ Processor.DirEntryForms + Processor.Separator + formName);
			FileInputStream fileIS = new FileInputStream(f);
			BufferedReader buf = new BufferedReader(new InputStreamReader(fileIS));
			String readString = new String();
			while ((readString = buf.readLine()) != null) {
				mDisplayForm += readString + "\n";
			}
			//Then two processings here: standard hard-coded forms and Custom forms
			if (!isCustomForm) {//Hard coded forms format 
				//Perform form specific processing
				if (formName.startsWith("radiogram.")) {
					//Reformat the boolean variables
					if (dataString.indexOf(":std:1 T") != -1) dataString = dataString.replace(":std:1 T", ":std:2 ON");
					if (dataString.indexOf(":std:1 F") != -1) dataString = dataString.replace(":std:1 F", ":std:3 OFF");
					if (dataString.indexOf(":svc:1 T") != -1) dataString = dataString.replace(":svc:1 T", ":svc:2 ON");
					if (dataString.indexOf(":svc:1 F") != -1) dataString = dataString.replace(":svc:1 F", ":svc:3 OFF");
				} else if (formName.startsWith("hics213.")) {
					//Reformat the boolean variables
					if (dataString.indexOf(":5a:1 X") != -1) dataString = dataString.replace(":5a:1 X", ":5.a:7 CHECKED");
					if (dataString.indexOf(":5b:1 X") != -1) dataString = dataString.replace(":5b:1 X", ":5.b:7 CHECKED");
					if (dataString.indexOf(":5c:1 X") != -1) dataString = dataString.replace(":5c:1 X", ":5.c:7 CHECKED");
					if (dataString.indexOf(":6a:1 X") != -1) dataString = dataString.replace(":6a:1 X", ":6.a:7 CHECKED");
					if (dataString.indexOf(":6b:1 X") != -1) dataString = dataString.replace(":6b:1 X", ":6.b:7 CHECKED");
					if (dataString.indexOf(":7a:1 X") != -1) dataString = dataString.replace(":7a:1 X", ":7.a:7 CHECKED");
					if (dataString.indexOf(":7b:1 X") != -1) dataString = dataString.replace(":7b:1 X", ":7.b:7 CHECKED");
					if (dataString.indexOf(":7c:1 X") != -1) dataString = dataString.replace(":7c:1 X", ":7.c:7 CHECKED");
				}
				//Re-format hard coded form format to custom form format
				Pattern psc = Pattern.compile("^(:.+:)(\\d+) (.*)|(^\\n)|(.+)", Pattern.MULTILINE);
				Matcher msc = psc.matcher(dataString);
				boolean keepLooking = true;
				String element;
				String value = "";
				String tempBuffer = "";
				int fieldLength = 0;
				int runningFieldLength = 0;
				boolean newUncompleteField = false;

				for (int start = 0; (keepLooking && start < dataString.length());)
				{
					keepLooking = msc.find(start);
					if (keepLooking) {
						if (msc.group(1) != null) { //We have a ":fieldname:nn fieldvalue" line
							newUncompleteField = true;//Assumed true, reset below if not true
							element = msc.group(1);
							// Processor.PostToTerminal("Field name <" + fieldName + ">\n");
							fieldLength = Integer.parseInt(msc.group(2));
							value = msc.group(3);
							// Processor.PostToTerminal("Field value <" + fieldValue + ">\n\n");
							if (fieldLength <= value.length()) {
								newUncompleteField = false;
								tempBuffer += element + "," + value + "\n";
							} else {
								runningFieldLength = value.length();
								tempBuffer += element + ",\"" + value;
							}
						} else if (msc.group(4) != null) {//We have just a line feed, nothing else
							if (newUncompleteField) {
								String a4 = msc.group(4);
								runningFieldLength += a4.length();
								tempBuffer += a4;
								if (fieldLength <= runningFieldLength) {
									newUncompleteField = false;
									tempBuffer += "\"\n";
								}
							}
						} else if (msc.group(5) != null){ //We have a 2nd, 3rd etc.. line of text
							if (newUncompleteField) {
								value = msc.group(5);
								runningFieldLength += value.length() + 1;
								tempBuffer += "\n" + value;
								if (fieldLength <= runningFieldLength) {
									newUncompleteField = false;
									tempBuffer += "\"\n";
								}
							}
						}
					}
					if (keepLooking) start = msc.end();
				}
				dataString = tempBuffer;
			}
			//Call Native code to format in HTML with updated VALUE and ReadOnly attributes
			mDisplayForm = Message.customHtmlEditingFormat(mDisplayForm, dataString); //C++ code
			//Debug
			//Processor.PostToTerminal("\nReturned from C++ <" + mDisplayForm + ">\n\n");
		}
		catch (FileNotFoundException e)
		{
			AndFlmsg.middleToastText("Form file " + formName + " Not Found. Can't Edit this message");
			e.printStackTrace();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		return mDisplayForm;
	}

	
	
	//Takes a directory for the data file, a data file name
	//Returns a wrapped and possibly compressed file ready for Txing via Radio Modems
	public static String formatForTx(String mDir, String mFileName) {
		String dataString = new String();
		String txBuffer = new String();
		String readString0 = new String();
		String headerString = new String();
		String chksumString = new String();
		boolean found_flmsg = false;
		String headerForm= new String();
		
		try
		{
			//First separate the header from the data fields (Headers are never compressed)
			//For this read the data file until we find the form information
			File fi = new File(Processor.HomePath + Processor.Dirprefix
					+ mDir + Processor.Separator + mFileName);
			FileInputStream fileISi = new FileInputStream(fi);
			BufferedReader buf0 = new BufferedReader(new InputStreamReader(fileISi));
			dataString = "";
			//Handles both hard-coded forms and custom forms
			while ((readString0 = buf0.readLine()) != null)
			{
				//While we are reading the file, store its complete content in a string for future usage
				dataString += readString0 + "\n";
				if (found_flmsg && readString0.contains(">") && readString0.startsWith("<")) {
					headerForm = readString0;
					headerString = dataString;
					//Reset for capturing the data fields values
					dataString = "";
				}
				if (readString0.startsWith("<flmsg>")) {
					found_flmsg = true;
				} 
			}
			//Include error handling if form not valid or does not include form name
			if (!found_flmsg){
				AndFlmsg.middleToastText("Missing <flmsg> statement at start of Message file. Is it corrupted?");
				return "";
			}
			Message.saveEnv(); //To enable access to preferences
			//Update header
			headerString = Message.updateHeader(headerString, AndFlmsg.header_reason_FROM, headerForm);
			//Process message body
			dataString = Message.compressMaybe(dataString);
			txBuffer = "[WRAP:fn " + mFileName + "]" + 
					headerString + dataString;
			chksumString = CheckSum.Crc16(txBuffer);
			txBuffer = "[WRAP:beg][WRAP:lf]" + txBuffer + 
			"[WRAP:chksum " + chksumString + "][WRAP:end]\n";
		}
		catch (FileNotFoundException e)
		{
			AndFlmsg.middleToastText("Message file " + formName + " Not Found. Deleted?");
			//e.printStackTrace();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		return txBuffer;
	}

	
	
	// Creates an array of flmsg messages file names
	public static String[] createFileListFromFolder(String mFolder)
	{
		String[] fileNamesArray = null;
		try
		{
			// Get the list of files in the designated folder
			File dir = new File(Processor.HomePath + Processor.Dirprefix
					+ mFolder);
			File[] files = dir.listFiles();
			FileFilter fileFilter = new FileFilter() {
				public boolean accept(File file)
				{
					return file.isFile();
				}
			};

			// We should now have an array of strings containing the file names
			files = dir.listFiles(fileFilter);
			if (files.length > 0) {
				fileNamesArray = new String[files.length] ;
				for (int i = 0; i < files.length; i++)
				{
					fileNamesArray[i] = files[i].getName();
				}
			}
		}
		catch (Exception e)
		{
			loggingclass.writelog("Error when listing Folder: " + mFolder + "\nDetails: ", e, true);
		}
		return fileNamesArray;
	}


	
	// Loads the list of Received flmsg messages
	public static void loadFileListFromFolder(String mFolder)
	{
		try
		{
			// Get the list of files in the designated folder
			File dir = new File(Processor.HomePath + Processor.Dirprefix
					+ mFolder);
			File[] files = dir.listFiles();
			FileFilter fileFilter = new FileFilter() {
				public boolean accept(File file)
				{
					return file.isFile();
				}
			};

			// We should now have an array of strings containing the file names
			files = dir.listFiles(fileFilter);
			AndFlmsg.msgArrayAdapter.clear();
			for (int i = 0; i < files.length; i++)
			{
				AndFlmsg.msgArrayAdapter.add(files[i].getName());
			}
		}
		catch (Exception e)
		{
			loggingclass.writelog("Error when showing Folder: " + mFolder + "\nDetails: ", e, true);
		}
	}


	public static String getExtension(String myFormName, boolean isTemplate) {
		String msgExtension = ".k2s";
		String tempExtension = ".k2t";

		if (myFormName.toLowerCase(Locale.US).startsWith("radiogram.")) {
			msgExtension = ".m2s";
			tempExtension = ".m2t";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("iaru.")) {
			msgExtension = ".i2s";
			tempExtension = ".i2t";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("plaintext.")) {
			msgExtension = ".p2s";
			tempExtension = ".p2t";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics203.")) {
			msgExtension = ".203";
			tempExtension = ".203T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics205.")) {
			msgExtension = ".205";
			tempExtension = ".205T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics205a.")) {
			msgExtension = ".25A";
			tempExtension = ".25AT";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics206.")) {
			msgExtension = ".206";
			tempExtension = ".206T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics213.")) {
			msgExtension = ".213";
			tempExtension = ".213T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics214.")) {
			msgExtension = ".214";
			tempExtension = ".214T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics216.")) {
			msgExtension = ".216";
			tempExtension = ".216T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("ics309.")) {
			msgExtension = ".309";
			tempExtension = ".309T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("hics203.")) {
			msgExtension = ".H203";
			tempExtension = ".H203T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("hics206.")) {
			msgExtension = ".H206";
			tempExtension = ".H206T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("hics213.")) {
			msgExtension = ".H213";
			tempExtension = ".H213T";
		} else if (myFormName.toLowerCase(Locale.US).startsWith("hics214.")) {
			msgExtension = ".H214";
			tempExtension = ".H214T";
		}
		//Return the value
		if (isTemplate) {
			return tempExtension;
		} else {
			return msgExtension;
		}
	}


	public static boolean isCustomForm(String extension) {
		if (extension.equals(".k2s") || extension.equals(".k2t")) {
			return true;
		} else return false;
	}
	
	
	
	//Extract the field content of a Hard Coded form buffer
	//The field name is passed without ":"
	public static String getHcfFieldContent(String dataBuffer, String fieldName) {
		String fieldValue = "";

		int posStart = dataBuffer.indexOf(":" + fieldName + ":");
		if (posStart != -1) { //There is Match
			//Get length of field
			int posBlankAfterLength = dataBuffer.indexOf(" ", posStart + fieldName.length() + 1);
			int countStringlength = posBlankAfterLength - (posStart + fieldName.length() + 2);
			if (posBlankAfterLength != -1 && countStringlength > 0) {
				int length;
				try {
					String lengthStr = dataBuffer.substring(posStart + fieldName.length() + 2, posBlankAfterLength);
					length = Integer.parseInt(lengthStr);
				} catch (NumberFormatException e) {
					length = 0;
				}
				//Extract value
				if (length > 0 && (dataBuffer.length() >= posBlankAfterLength + length + 1)) {
					fieldValue = dataBuffer.substring(posBlankAfterLength + 1, length + posBlankAfterLength + 1);
				}
			}
		}
		return fieldValue;
	}
	
	
	
	//Set the field content of a Hard Coded form buffer
	public static String setHcfFieldContent(String dataBuffer, String fieldName, String newValue) {
		String newBuffer = "";

		int posStart = dataBuffer.indexOf(":" + fieldName + ":");
		if (posStart != -1) { //There is Match
			//Get length of field
			int posBlankAfterLength = dataBuffer.indexOf(" ", posStart + fieldName.length() + 1);
			int countStringlength = posBlankAfterLength - (posStart + fieldName.length() + 2);
			if (posBlankAfterLength != -1 && countStringlength > 0) {
				int length;
				try {
					String lengthStr = dataBuffer.substring(posStart + fieldName.length() + 2, posBlankAfterLength);
					length = Integer.parseInt(lengthStr);
				} catch (NumberFormatException e) {
					length = 0;
				}
				//Build section to be replaced (field, length and value)
				String toBeReplaced = dataBuffer.substring(posStart, length + posBlankAfterLength + 1);
				//Build new replacement section. If newValue is blank, remove the whole entry
				String replacement = "";
				if (!newValue.equals("")) {
					replacement = ":" + fieldName + ":" + Integer.toString(newValue.length()) + " " + newValue;
				}
				newBuffer = dataBuffer.replace(toBeReplaced, replacement);
			}
		} else { //add field since it does not exist in buffer (only if newValue is not blank)
			if (!newValue.equals("")) {
				newBuffer = dataBuffer + ":" + fieldName + ":" + Integer.toString(newValue.length()) + " " + newValue + "\n";
			}
		}
		return newBuffer;
	}
	
	
}
